/**
 * 
 * @author User MightyPork from stackoverflow.com
 * Modifications by Dr. Vargas-Perez, Fall 2018
 * 
 * This class creates and prints a visual Binary Tree when the root of
 * that tree is given.
 * 
 * This class will always print a perfect binary tree, meaning that the 
 * empty nodes are going to be completed with null.
 *
 */

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

public class PrintBinaryTree {
	
	AVLNode root;
	
	public PrintBinaryTree(AVLNode root) {
		this.root = root;
	}
	
	/**
	 * This method performs a level-order traversal of the tree
	 * to get all the nodes level by level.
	 */
	public void printTree() {
		Queue<AVLNode> levelQueue = new LinkedList<AVLNode>();
		List<List<String>> nodes = new ArrayList<List<String>>();
		int maxWidth = 0;
		
		levelQueue.add(root);
		while(!levelQueue.isEmpty())
		{
			List<String> node = new ArrayList<String>();
			int numNodes = levelQueue.size();
			boolean allNull = true;
			
			while(numNodes > 0) {
				
				AVLNode parent = levelQueue.remove();
				numNodes--;
				
				if (parent != null)
				{
					String e = String.valueOf(parent.getElement());
					node.add(e);
					
					if (e.length() > maxWidth) 
						maxWidth = e.length();
					
					levelQueue.add(parent.getLeftChild());
					levelQueue.add(parent.getRightChild());
					
					if(parent.getLeftChild() != parent.getRightChild()) 
						allNull = false;	
				}
				else {
					node.add("•");
					levelQueue.add(null);
					levelQueue.add(null);
				}
			}
			
			if (maxWidth % 2 == 1) 
				maxWidth++;

			nodes.add(node);
			
			// If all the nodes in the next level are null
			if(allNull)
				break;			
		}
		
		buildTree(nodes, maxWidth);
	}
	
	/**
	 * This is a private helper method that takes the number of tree levels
	 * and nodes per levels, as well as the maximum length of nodes
	 * and builds (prints) the tree diagram.  
	 * 
	 * @param levels
	 * @param maxWidth
	 */
	private void buildTree(List<List<String>> levels, int maxWidth) {
		int maxSize = levels.get(levels.size() - 1).size() * (maxWidth + 2);
		
		for (int i = 0; i < levels.size(); i++) {
			List<String> nodes = levels.get(i);

			// This section prints "branches" of the tree
			if (i > 0) {
				int hpw = (int) Math.floor(maxSize / 2) - 1;
				for (int j = 0; j < nodes.size(); j++) {

					// split node
					char c = ' ';
					if (j % 2 == 1) {
						c = '┴';
					}
					System.out.print(c);

					// lines and spaces
					for (int k = 0; k < hpw; k++) {
						System.out.print(j % 2 == 0 ? " " : "─");
					}
					System.out.print(j % 2 == 0 ? "┌" : "┐");
					for (int k = 0; k < hpw; k++) {
						System.out.print(j % 2 == 0 ? "─" : " ");
					}
				}
				System.out.println();
			}

			// print line of numbers
			for (int j = 0; j < nodes.size(); j++) {

				String e = nodes.get(j);
				
				int leftMargin = (int) Math.ceil(maxSize / 2f - e.length() / 2f);
				int rightMargin = (int) Math.floor(maxSize / 2f - e.length() / 2f);

				// a number
				for (int k = 0; k < leftMargin; k++) {
					System.out.print(" ");
				}
				System.out.print(e);
				for (int k = 0; k < rightMargin; k++) {
					System.out.print(" ");
				}
			}
			System.out.println();

			maxSize /= 2;
		}
	}
}
